'use strict';

Object.defineProperty(exports, '__esModule', { value: true });

var priorityQueue = require('@algorithm.ts/priority-queue');

exports.GomokuDirectionType = void 0;
(function (GomokuDirectionType) {
    GomokuDirectionType[GomokuDirectionType["LEFT"] = 0] = "LEFT";
    GomokuDirectionType[GomokuDirectionType["TOP_LEFT"] = 2] = "TOP_LEFT";
    GomokuDirectionType[GomokuDirectionType["TOP"] = 4] = "TOP";
    GomokuDirectionType[GomokuDirectionType["TOP_RIGHT"] = 6] = "TOP_RIGHT";
    GomokuDirectionType[GomokuDirectionType["RIGHT"] = 1] = "RIGHT";
    GomokuDirectionType[GomokuDirectionType["BOTTOM_RIGHT"] = 3] = "BOTTOM_RIGHT";
    GomokuDirectionType[GomokuDirectionType["BOTTOM"] = 5] = "BOTTOM";
    GomokuDirectionType[GomokuDirectionType["BOTTOM_LEFT"] = 7] = "BOTTOM_LEFT";
})(exports.GomokuDirectionType || (exports.GomokuDirectionType = {}));
const GomokuDirections = Array.from(Object.entries({
    [exports.GomokuDirectionType.TOP]: [-1, 0],
    [exports.GomokuDirectionType.TOP_RIGHT]: [-1, 1],
    [exports.GomokuDirectionType.RIGHT]: [0, 1],
    [exports.GomokuDirectionType.BOTTOM_RIGHT]: [1, 1],
    [exports.GomokuDirectionType.BOTTOM]: [1, 0],
    [exports.GomokuDirectionType.BOTTOM_LEFT]: [1, -1],
    [exports.GomokuDirectionType.LEFT]: [0, -1],
    [exports.GomokuDirectionType.TOP_LEFT]: [-1, -1],
}).reduce((acc, [key, value]) => {
    acc[key] = value;
    return acc;
}, []));
const GomokuDirectionTypes = {
    full: [
        exports.GomokuDirectionType.LEFT,
        exports.GomokuDirectionType.TOP_LEFT,
        exports.GomokuDirectionType.TOP,
        exports.GomokuDirectionType.TOP_RIGHT,
        exports.GomokuDirectionType.RIGHT,
        exports.GomokuDirectionType.BOTTOM_RIGHT,
        exports.GomokuDirectionType.BOTTOM,
        exports.GomokuDirectionType.BOTTOM_LEFT,
    ],
    leftHalf: [
        exports.GomokuDirectionType.LEFT,
        exports.GomokuDirectionType.TOP_LEFT,
        exports.GomokuDirectionType.TOP,
        exports.GomokuDirectionType.TOP_RIGHT,
    ],
    rightHalf: [
        exports.GomokuDirectionType.RIGHT,
        exports.GomokuDirectionType.BOTTOM_RIGHT,
        exports.GomokuDirectionType.BOTTOM,
        exports.GomokuDirectionType.BOTTOM_LEFT,
    ],
};
const GomokuDirectionTypeBitset = {
    full: GomokuDirectionTypes.full.reduce((acc, dirType) => acc | (1 << dirType), 0),
    leftHalf: GomokuDirectionTypes.leftHalf.reduce((acc, dirType) => acc | (1 << dirType), 0),
    rightHalf: GomokuDirectionTypes.rightHalf.reduce((acc, dirType) => acc | (1 << dirType), 0),
};

const { full: fullDirectionTypes$2 } = GomokuDirectionTypes;
class GomokuContext {
    MAX_ROW;
    MAX_COL;
    MAX_ADJACENT;
    MAX_DISTANCE_OF_NEIGHBOR;
    TOTAL_POS;
    MIDDLE_POS;
    board;
    idx;
    _idxMap;
    _gomokuDirections;
    _maxMovableMap;
    _dirStartPosMap;
    _dirStartPosSet;
    _dirNeighborSet;
    _neighborPlacedCount;
    _placedCount;
    constructor(props) {
        const { MAX_ROW, MAX_COL, MAX_ADJACENT, MAX_DISTANCE_OF_NEIGHBOR } = props;
        const _MAX_ROW = Math.max(1, MAX_ROW);
        const _MAX_COL = Math.max(1, MAX_COL);
        const _MAX_ADJACENT = Math.max(1, MAX_ADJACENT);
        const _MAX_DISTANCE_OF_NEIGHBOR = Math.max(1, MAX_DISTANCE_OF_NEIGHBOR);
        const _TOTAL_POS = _MAX_ROW * _MAX_COL;
        const idx = (r, c) => r * _MAX_ROW + c;
        this.MAX_ROW = _MAX_ROW;
        this.MAX_COL = _MAX_COL;
        this.MAX_ADJACENT = _MAX_ADJACENT;
        this.MAX_DISTANCE_OF_NEIGHBOR = _MAX_DISTANCE_OF_NEIGHBOR;
        this.TOTAL_POS = _TOTAL_POS;
        this.MIDDLE_POS = _TOTAL_POS >> 1;
        this.board = new Array(_TOTAL_POS).fill(-1);
        this.idx = idx;
        const _idxMap = new Array(_TOTAL_POS);
        for (let r = 0; r < _MAX_ROW; ++r) {
            for (let c = 0; c < _MAX_COL; ++c) {
                const posId = idx(r, c);
                _idxMap[posId] = [r, c];
            }
        }
        const _gomokuDirections = GomokuDirections.map(([dr, dc]) => dr * MAX_ROW + dc);
        const _maxMovableMap = new Array(fullDirectionTypes$2.length)
            .fill([])
            .map(() => new Array(_TOTAL_POS).fill(0));
        const _dirStartPosSet = new Array(fullDirectionTypes$2.length).fill([]).map(() => []);
        const _dirStartPosMap = new Array(fullDirectionTypes$2.length)
            .fill([])
            .map(() => new Array(_TOTAL_POS).fill(0));
        this.traverseAllDirections(dirType => {
            const revDirType = dirType ^ 1;
            const [dr, dc] = GomokuDirections[dirType];
            return posId => {
                const [r, c] = _idxMap[posId];
                const r2 = r + dr;
                const c2 = c + dc;
                if (r2 < 0 || r2 >= _MAX_ROW || c2 < 0 || c2 >= _MAX_ROW) {
                    _maxMovableMap[dirType][posId] = 0;
                    _dirStartPosMap[revDirType][posId] = posId;
                    _dirStartPosSet[revDirType].push(posId);
                }
                else {
                    const posId2 = idx(r2, c2);
                    _maxMovableMap[dirType][posId] = _maxMovableMap[dirType][posId2] + 1;
                    _dirStartPosMap[revDirType][posId] = _dirStartPosMap[revDirType][posId2];
                }
            };
        });
        const _neighborMap = new Array(_TOTAL_POS);
        for (let posId = 0; posId < _TOTAL_POS; ++posId) {
            const neighbors = [];
            _neighborMap[posId] = neighbors;
            for (const dirType of fullDirectionTypes$2) {
                let posId2 = posId;
                for (let step = 0; step < _MAX_DISTANCE_OF_NEIGHBOR; ++step) {
                    if (1 > _maxMovableMap[dirType][posId2])
                        break;
                    posId2 += _gomokuDirections[dirType];
                    neighbors.push(posId2);
                }
            }
        }
        this._gomokuDirections = _gomokuDirections;
        this._idxMap = _idxMap;
        this._maxMovableMap = _maxMovableMap;
        this._dirStartPosMap = _dirStartPosMap;
        this._dirStartPosSet = _dirStartPosSet;
        this._dirNeighborSet = _neighborMap;
        this._neighborPlacedCount = new Array(_TOTAL_POS).fill(0);
        this._placedCount = 0;
    }
    get placedCount() {
        return this._placedCount;
    }
    init(pieces) {
        const board = this.board;
        board.fill(-1);
        this._placedCount = 0;
        const { _neighborPlacedCount } = this;
        _neighborPlacedCount.fill(0);
        for (const { r, c, p } of pieces) {
            const posId = this.idx(r, c);
            if (board[posId] < 0) {
                board[posId] = p;
                this._placedCount += 1;
                for (const neighborId of this.accessibleNeighbors(posId)) {
                    _neighborPlacedCount[neighborId] += 1;
                }
            }
        }
    }
    forward(posId, playerId) {
        const board = this.board;
        if (posId < 0 || posId >= this.TOTAL_POS || board[posId] >= 0)
            return false;
        board[posId] = playerId;
        this._placedCount += 1;
        for (const neighborId of this.accessibleNeighbors(posId)) {
            this._neighborPlacedCount[neighborId] += 1;
        }
        return true;
    }
    revert(posId) {
        const board = this.board;
        if (posId < 0 || posId >= this.TOTAL_POS || board[posId] < 0)
            return false;
        board[posId] = -1;
        this._placedCount -= 1;
        for (const neighborId of this.accessibleNeighbors(posId)) {
            this._neighborPlacedCount[neighborId] -= 1;
        }
        return true;
    }
    revIdx(posId) {
        return this._idxMap[posId];
    }
    isValidPos(r, c) {
        return r >= 0 && r < this.MAX_ROW && c >= 0 && c < this.MAX_COL;
    }
    safeMove(posId, dirType, step) {
        return step <= this._maxMovableMap[dirType][posId]
            ? posId + this._gomokuDirections[dirType] * step
            : -1;
    }
    safeMoveOneStep(posId, dirType) {
        return 1 <= this._maxMovableMap[dirType][posId] ? posId + this._gomokuDirections[dirType] : -1;
    }
    fastMove(posId, dirType, step) {
        return posId + this._gomokuDirections[dirType] * step;
    }
    fastMoveOneStep(posId, dirType) {
        return posId + this._gomokuDirections[dirType];
    }
    maxMovableSteps(posId, dirType) {
        return this._maxMovableMap[dirType][posId];
    }
    accessibleNeighbors(posId) {
        return this._dirNeighborSet[posId];
    }
    hasPlacedNeighbors(posId) {
        return this._neighborPlacedCount[posId] > 0;
    }
    getStartPosId(posId, dirType) {
        return this._dirStartPosMap[dirType][posId];
    }
    getStartPosSet(dirType) {
        return this._dirStartPosSet[dirType];
    }
    traverseAllDirections(handle) {
        const { TOTAL_POS } = this;
        const { leftHalf, rightHalf } = GomokuDirectionTypes;
        for (const dirType of leftHalf) {
            const h = handle(dirType);
            for (let posId = 0; posId < TOTAL_POS; ++posId)
                h(posId);
        }
        for (const dirType of rightHalf) {
            const h = handle(dirType);
            for (let posId = TOTAL_POS - 1; posId >= 0; --posId)
                h(posId);
        }
    }
}

const { full: fullDirectionTypes$1, rightHalf: halfDirectionTypes$1 } = GomokuDirectionTypes;
class GomokuCountMap {
    context;
    _rightHalfDirCountMap;
    constructor(context) {
        this.context = context;
        this._rightHalfDirCountMap = new Array(fullDirectionTypes$1.length)
            .fill([])
            .map(() => new Array(context.TOTAL_POS).fill([]).map(() => []));
    }
    init() {
        const { context } = this;
        const { board } = context;
        const { _rightHalfDirCountMap } = this;
        for (const dirType of halfDirectionTypes$1) {
            for (const startPosId of context.getStartPosSet(dirType)) {
                const counters = _rightHalfDirCountMap[dirType][startPosId];
                let index = 0;
                const maxSteps = context.maxMovableSteps(startPosId, dirType) + 1;
                for (let i = 0, posId = startPosId, i2, posId2; i < maxSteps; i = i2, posId = posId2) {
                    const playerId = board[posId];
                    for (i2 = i + 1, posId2 = posId; i2 < maxSteps; ++i2) {
                        posId2 = context.fastMoveOneStep(posId2, dirType);
                        if (board[posId2] !== playerId)
                            break;
                    }
                    counters[index++] = { playerId, count: i2 - i };
                }
                counters.length = index;
            }
        }
    }
    forward(posId) {
        const playerId = this.context.board[posId];
        for (const dirType of halfDirectionTypes$1) {
            this._updateHalfDirCounter(playerId, posId, dirType);
        }
    }
    revert(posId) {
        for (const dirType of halfDirectionTypes$1) {
            this._updateHalfDirCounter(-1, posId, dirType);
        }
    }
    getDirCounters(startPosId, dirType) {
        return this._rightHalfDirCountMap[dirType][startPosId];
    }
    _updateHalfDirCounter(playerId, posId, dirType) {
        const { context } = this;
        const revDirType = dirType ^ 1;
        const startPosId = context.getStartPosId(posId, dirType);
        const counters = this._rightHalfDirCountMap[dirType][startPosId];
        let index = 0;
        let remain = context.maxMovableSteps(posId, revDirType) + 1;
        for (; index < counters.length; ++index) {
            const cnt = counters[index].count;
            if (cnt >= remain)
                break;
            remain -= cnt;
        }
        if (remain === 1) {
            if (counters[index].count === 1) {
                const isSameWithLeft = index > 0 && counters[index - 1].playerId === playerId;
                const isSameWithRight = index + 1 < counters.length && counters[index + 1].playerId === playerId;
                if (isSameWithLeft) {
                    if (isSameWithRight) {
                        counters[index - 1].count += 1 + counters[index + 1].count;
                        counters.splice(index, 2);
                    }
                    else {
                        counters[index - 1].count += 1;
                        counters.splice(index, 1);
                    }
                }
                else {
                    if (isSameWithRight) {
                        counters[index + 1].count += 1;
                        counters.splice(index, 1);
                    }
                    else {
                        counters[index].playerId = playerId;
                    }
                }
            }
            else {
                counters[index].count -= 1;
                if (index > 0 && counters[index - 1].playerId === playerId) {
                    counters[index - 1].count += 1;
                }
                else {
                    counters.splice(index, 0, { playerId, count: 1 });
                }
            }
        }
        else if (remain < counters[index].count) {
            const { playerId: hitPlayerId, count: hitCount } = counters[index];
            counters.splice(index, 1, { playerId: hitPlayerId, count: remain - 1 }, { playerId, count: 1 }, { playerId: hitPlayerId, count: hitCount - remain });
        }
        else {
            counters[index].count -= 1;
            if (index + 1 < counters.length && counters[index + 1].playerId === playerId) {
                counters[index + 1].count += 1;
            }
            else {
                counters.splice(index + 1, 0, { playerId, count: 1 });
            }
        }
    }
}

const { full: fullDirectionTypes, rightHalf: halfDirectionTypes } = GomokuDirectionTypes;
const { rightHalf: allDirectionTypeBitset } = GomokuDirectionTypeBitset;
class GomokuState {
    context;
    NEXT_MOVER_BUFFER = 2;
    _countMap;
    _scoreMap;
    _countOfReachLimits;
    _countOfReachLimitsDirMap;
    _stateScoreMap;
    _stateScoreDirMap;
    _candidateSet;
    _candidateScoreMap;
    _candidateScoreDirMap;
    _candidateScoreExpired;
    constructor(props) {
        const { context, scoreMap } = props;
        const _countOfReachLimitsDirMap = new Array(fullDirectionTypes.length)
            .fill([])
            .map(() => new Array(context.TOTAL_POS).fill([]).map(() => [0, 0]));
        const _stateScoreDirMap = new Array(fullDirectionTypes.length)
            .fill([])
            .map(() => new Array(context.TOTAL_POS).fill([]).map(() => [0, 0]));
        const _candidateScoreMap = new Array(context.TOTAL_POS)
            .fill([])
            .map(() => [0, 0]);
        const _candidateScoreDirMap = new Array(fullDirectionTypes.length)
            .fill([])
            .map(() => new Array(context.TOTAL_POS).fill([]).map(() => [0, 0]));
        const _candidateScoreExpired = new Array(context.TOTAL_POS).fill(allDirectionTypeBitset);
        this.context = context;
        this._countMap = new GomokuCountMap(context);
        this._scoreMap = scoreMap;
        this._countOfReachLimits = [0, 0];
        this._countOfReachLimitsDirMap = _countOfReachLimitsDirMap;
        this._stateScoreMap = [0, 0];
        this._stateScoreDirMap = _stateScoreDirMap;
        this._candidateSet = new Set();
        this._candidateScoreMap = _candidateScoreMap;
        this._candidateScoreDirMap = _candidateScoreDirMap;
        this._candidateScoreExpired = _candidateScoreExpired;
    }
    init(pieces) {
        const { context, _countOfReachLimits, _countOfReachLimitsDirMap, _stateScoreMap, _stateScoreDirMap, _candidateSet, _candidateScoreExpired, } = this;
        this._countMap.init();
        _countOfReachLimits.fill(0);
        _stateScoreMap.fill(0);
        for (const dirType of halfDirectionTypes) {
            for (const startPosId of context.getStartPosSet(dirType)) {
                const { scores, countOfReachLimits } = this._evaluateScoreInDirection(startPosId, dirType);
                _countOfReachLimits[0] += countOfReachLimits[0];
                _countOfReachLimits[1] += countOfReachLimits[1];
                _countOfReachLimitsDirMap[dirType][startPosId][0] = countOfReachLimits[0];
                _countOfReachLimitsDirMap[dirType][startPosId][1] = countOfReachLimits[1];
                _stateScoreMap[0] += scores[0];
                _stateScoreMap[1] += scores[1];
                _stateScoreDirMap[dirType][startPosId][0] = scores[0];
                _stateScoreDirMap[dirType][startPosId][1] = scores[1];
            }
        }
        _candidateSet.clear();
        for (const { r, c } of pieces) {
            const posId = context.idx(r, c);
            for (const posId2 of context.accessibleNeighbors(posId)) {
                if (context.board[posId2] < 0)
                    _candidateSet.add(posId2);
            }
        }
        for (let posId = 0; posId < context.TOTAL_POS; ++posId) {
            _candidateScoreExpired[posId] = allDirectionTypeBitset;
        }
    }
    forward(posId) {
        this._countMap.forward(posId);
        const { context, _candidateSet } = this;
        _candidateSet.delete(posId);
        for (const posId2 of context.accessibleNeighbors(posId)) {
            if (context.board[posId2] < 0)
                _candidateSet.add(posId2);
        }
        this._updateStateScore(posId);
        this._expireCandidates(posId);
    }
    revert(posId) {
        this._countMap.revert(posId);
        const { context, _candidateSet } = this;
        if (context.hasPlacedNeighbors(posId))
            _candidateSet.add(posId);
        for (const posId2 of context.accessibleNeighbors(posId)) {
            if (!context.hasPlacedNeighbors(posId2))
                _candidateSet.delete(posId2);
        }
        this._updateStateScore(posId);
        this._expireCandidates(posId);
    }
    expand(nextPlayerId, candidates) {
        const { context, _candidateSet } = this;
        if (context.board[context.MIDDLE_POS] < 0)
            _candidateSet.add(this.context.MIDDLE_POS);
        let i = 0;
        for (const posId of _candidateSet) {
            const score = this._evaluateCandidate(posId, nextPlayerId);
            candidates[i] = { id: posId, score };
            i += 1;
        }
        return _candidateSet.size;
    }
    score(currentPlayer, scoreForPlayer) {
        const score0 = this._stateScoreMap[scoreForPlayer];
        const score1 = this._stateScoreMap[scoreForPlayer ^ 1];
        const nextMoverFac = this.NEXT_MOVER_BUFFER;
        return currentPlayer === scoreForPlayer
            ? score0 - score1 * nextMoverFac
            : score0 * nextMoverFac - score1;
    }
    isWin(currentPlayer) {
        return this._countOfReachLimits[currentPlayer] > 0;
    }
    isDraw() {
        return this.context.placedCount === this.context.TOTAL_POS;
    }
    isFinal() {
        const { context, _countOfReachLimits } = this;
        if (context.placedCount === context.TOTAL_POS)
            return true;
        if (_countOfReachLimits[0] > 0 || _countOfReachLimits[1] > 0)
            return true;
        return false;
    }
    _updateStateScore(posId) {
        const { context, _countOfReachLimits, _countOfReachLimitsDirMap, _stateScoreMap, _stateScoreDirMap, } = this;
        for (const dirType of halfDirectionTypes) {
            const startPosId = context.getStartPosId(posId, dirType);
            const { scores: [score0, score1], countOfReachLimits: [count0, count1], } = this._evaluateScoreInDirection(startPosId, dirType);
            _countOfReachLimits[0] += count0 - _countOfReachLimitsDirMap[dirType][startPosId][0];
            _countOfReachLimits[1] += count1 - _countOfReachLimitsDirMap[dirType][startPosId][1];
            _countOfReachLimitsDirMap[dirType][startPosId][0] = count0;
            _countOfReachLimitsDirMap[dirType][startPosId][1] = count1;
            _stateScoreMap[0] += score0 - _stateScoreDirMap[dirType][startPosId][0];
            _stateScoreMap[1] += score1 - _stateScoreDirMap[dirType][startPosId][1];
            _stateScoreDirMap[dirType][startPosId][0] = score0;
            _stateScoreDirMap[dirType][startPosId][1] = score1;
        }
    }
    _expireCandidates(posId) {
        const { context, _candidateScoreExpired } = this;
        for (const dirType of halfDirectionTypes) {
            const startPosId = context.getStartPosId(posId, dirType);
            const maxSteps = context.maxMovableSteps(startPosId, dirType);
            const bitMark = 1 << dirType;
            for (let id = startPosId, step = 0;; ++step) {
                _candidateScoreExpired[id] |= bitMark;
                if (step === maxSteps)
                    break;
                id = context.fastMoveOneStep(id, dirType);
            }
        }
    }
    _evaluateCandidate(posId, nextPlayerId) {
        const { _candidateScoreExpired, _candidateScoreMap } = this;
        const expiredBitset = _candidateScoreExpired[posId];
        if (expiredBitset > 0) {
            const { context, _candidateScoreDirMap, _stateScoreDirMap } = this;
            let prevScore0 = 0;
            let prevScore1 = 0;
            for (const dirType of halfDirectionTypes) {
                const startPosId = context.getStartPosId(posId, dirType);
                prevScore0 += _stateScoreDirMap[dirType][startPosId][0];
                prevScore1 += _stateScoreDirMap[dirType][startPosId][1];
            }
            let score0 = 0;
            this._forward(posId, 0);
            for (const dirType of halfDirectionTypes) {
                if ((1 << dirType) & expiredBitset) {
                    const startPosId = context.getStartPosId(posId, dirType);
                    const { scores } = this._evaluateScoreInDirection(startPosId, dirType);
                    _candidateScoreDirMap[dirType][posId][0] = scores[0];
                    score0 += scores[0];
                }
                else {
                    score0 += _candidateScoreDirMap[dirType][posId][0];
                }
            }
            this._revert(posId);
            let score1 = 0;
            this._forward(posId, 1);
            for (const dirType of halfDirectionTypes) {
                if ((1 << dirType) & expiredBitset) {
                    const startPosId = context.getStartPosId(posId, dirType);
                    const { scores } = this._evaluateScoreInDirection(startPosId, dirType);
                    _candidateScoreDirMap[dirType][posId][1] = scores[1];
                    score1 += scores[1];
                }
                else {
                    score1 += _candidateScoreDirMap[dirType][posId][1];
                }
            }
            this._revert(posId);
            const deltaScore0 = score0 - prevScore0;
            const deltaScore1 = score1 - prevScore1;
            const baseScore = deltaScore0 + deltaScore1;
            _candidateScoreMap[posId][0] = baseScore + deltaScore0;
            _candidateScoreMap[posId][1] = baseScore + deltaScore1;
            _candidateScoreExpired[posId] = 0;
        }
        return _candidateScoreMap[posId][nextPlayerId];
    }
    _evaluateScoreInDirection(startPosId, dirType) {
        const { context } = this;
        const { con, gap } = this._scoreMap;
        const { MAX_ADJACENT } = context;
        const THRESHOLD = MAX_ADJACENT - 1;
        const counters = this._countMap.getDirCounters(startPosId, dirType);
        const _size = counters.length;
        const scores = [0, 0];
        const countOfReachLimits = [0, 0];
        for (let i = 0, j, k; i < _size; i = j) {
            if (i > 0 && counters[i - 1].playerId < 0)
                i -= 1;
            k = i;
            j = i;
            let { playerId, count: maxPossibleCnt } = counters[j];
            if (playerId < 0) {
                if (++j === _size)
                    break;
                const { playerId: playerId2, count: cnt } = counters[j];
                playerId = playerId2;
                maxPossibleCnt += cnt;
                k += 1;
            }
            for (j += 1; j < _size; ++j) {
                const { playerId: playerId2, count: cnt } = counters[j];
                if (playerId2 >= 0 && playerId2 !== playerId)
                    break;
                maxPossibleCnt += cnt;
            }
            if (maxPossibleCnt >= MAX_ADJACENT) {
                let isFreeSide0 = k > i ? 1 : 0;
                for (let usedK = -1; k < j; k += 2, isFreeSide0 = 1) {
                    const cnt1 = counters[k].count;
                    if (cnt1 < THRESHOLD && k + 2 < j && counters[k + 1].count === 1) {
                        const cnt2 = counters[k + 2].count;
                        if (cnt2 < THRESHOLD) {
                            const isFreeSide2 = k + 3 < j ? 1 : 0;
                            const normalizedCnt = Math.min(cnt1 + cnt2, THRESHOLD);
                            scores[playerId] += gap[normalizedCnt][isFreeSide0 + isFreeSide2];
                            usedK = k + 2;
                            continue;
                        }
                    }
                    if (usedK !== k) {
                        let normalizedCnt = cnt1;
                        if (cnt1 >= MAX_ADJACENT) {
                            normalizedCnt = MAX_ADJACENT;
                            countOfReachLimits[playerId] += 1;
                        }
                        const isFreeSide2 = k + 1 < j ? 1 : 0;
                        scores[playerId] += con[normalizedCnt][isFreeSide0 + isFreeSide2];
                    }
                }
            }
        }
        return { scores, countOfReachLimits };
    }
    _forward(posId, playerId) {
        this.context.forward(posId, playerId);
        this._countMap.forward(posId);
    }
    _revert(posId) {
        this.context.revert(posId);
        this._countMap.revert(posId);
    }
}

class GomokuStateCompressor {
    MOVE_STEP_BIT_BASE;
    MOVE_STEP_MASK;
    MOVE_STEP_MINIMUM_FLAG;
    INITIAL_STATE = 1n;
    constructor(TOTAL_POS) {
        let COMPRESS_MOVE_BIT_BASE = 1n;
        while (1n << COMPRESS_MOVE_BIT_BASE < TOTAL_POS)
            COMPRESS_MOVE_BIT_BASE += 1n;
        this.MOVE_STEP_BIT_BASE = COMPRESS_MOVE_BIT_BASE;
        this.MOVE_STEP_MASK = (1n << COMPRESS_MOVE_BIT_BASE) - 1n;
        this.MOVE_STEP_MINIMUM_FLAG = 1n << COMPRESS_MOVE_BIT_BASE;
    }
    compress(countInState, state, newMoveStepId) {
        if (countInState === 0)
            return newMoveStepId + this.MOVE_STEP_MINIMUM_FLAG;
        if (countInState === 1)
            return (state << this.MOVE_STEP_BIT_BASE) | newMoveStepId;
        const rhtCnt = countInState >> 1;
        const RIGHT_BITS = this.MOVE_STEP_BIT_BASE * BigInt(rhtCnt);
        const rht = state & ((1n << RIGHT_BITS) - 1n);
        if (countInState & 1) {
            const rhtMoveIds = this.decompressMoveIds(rhtCnt, rht, newMoveStepId);
            const newRht = this.compressMoveIds(rhtMoveIds);
            return ((state ^ rht) << this.MOVE_STEP_BIT_BASE) | newRht;
        }
        else {
            const lftCnt = countInState - rhtCnt;
            const lft = state >> RIGHT_BITS;
            const lftMoveIds = this.decompressMoveIds(lftCnt, lft, newMoveStepId);
            const NEW_LEFT_BITS = this.MOVE_STEP_BIT_BASE * BigInt(lftCnt + 1);
            const newLft = this.compressMoveIds(lftMoveIds) | (1n << NEW_LEFT_BITS);
            return (newLft << RIGHT_BITS) | rht;
        }
    }
    decompressMoveIds(cnt, state, newMoveStepId) {
        const { MOVE_STEP_BIT_BASE, MOVE_STEP_MASK } = this;
        const moveIds = [];
        let i = 0;
        let v = state;
        let id = 0n;
        for (; i < cnt; ++i) {
            id = v & MOVE_STEP_MASK;
            v >>= MOVE_STEP_BIT_BASE;
            if (id < newMoveStepId)
                break;
            moveIds.push(id);
        }
        moveIds.push(newMoveStepId);
        if (i < cnt)
            moveIds.push(id);
        for (i += 1; i < cnt; ++i) {
            id = v & MOVE_STEP_MASK;
            v >>= MOVE_STEP_BIT_BASE;
            moveIds.push(id);
        }
        return moveIds.reverse();
    }
    compressMoveIds(moveIds) {
        const { MOVE_STEP_BIT_BASE } = this;
        let result = 0n;
        for (const id of moveIds) {
            result = (result << MOVE_STEP_BIT_BASE) | id;
        }
        return result;
    }
}

class GomokuStateCache {
    MAX_STATES;
    stateCompressor;
    states;
    constructor(TOTAL_POS, MAX_STATES = 2e5) {
        this.stateCompressor = new GomokuStateCompressor(TOTAL_POS);
        this.states = new Map();
        this.MAX_STATES = MAX_STATES;
    }
    get INITIAL_STATE() {
        return this.stateCompressor.INITIAL_STATE;
    }
    clear() {
        this.states.clear();
    }
    get(state) {
        return this.states.get(state);
    }
    set(state, score) {
        if (this.states.size < this.MAX_STATES)
            this.states.set(state, score);
    }
    calcNextState(cur, prevState, nextMovePosId) {
        const nextState = this.stateCompressor.compress(cur, prevState, BigInt(nextMovePosId));
        return nextState;
    }
}

const createScoreMap = (MAX_ADJACENT) => {
    const con = new Array(MAX_ADJACENT + 1).fill([]).map(() => [0, 0, 0]);
    const gap = new Array(MAX_ADJACENT + 1).fill([]).map(() => [0, 0, 0]);
    let BASE_VALUE = 2;
    for (let cnt = 1; cnt < MAX_ADJACENT; ++cnt, BASE_VALUE *= 16) {
        con[cnt] = [0, BASE_VALUE, BASE_VALUE * 2];
        con[cnt] = [0, BASE_VALUE / 2, BASE_VALUE];
    }
    const _v = con[MAX_ADJACENT - 1][1];
    gap[MAX_ADJACENT - 1] = [_v, _v, _v];
    con[MAX_ADJACENT] = [_v, _v, _v];
    con[MAX_ADJACENT] = [BASE_VALUE, BASE_VALUE, BASE_VALUE];
    return { con, gap };
};

const tmpCandidates = [];
class GomokuSolution {
    MIN_DEPTH;
    MAX_DEPTH;
    POSSIBILITY_SEARCH_EQUIV_CANDIDATE;
    context;
    state;
    _cache;
    _queues;
    _bestMoveId;
    _mainPlayerId;
    constructor(props) {
        const { MAX_ROW, MAX_COL, MAX_ADJACENT = 5, MIN_DEPTH = 3, MAX_DEPTH = 7, MAX_DISTANCE_OF_NEIGHBOR = 2, POSSIBILITY_SEARCH_EQUIV_CANDIDATE = 0.98, } = props;
        const _context = new GomokuContext({ MAX_ROW, MAX_COL, MAX_ADJACENT, MAX_DISTANCE_OF_NEIGHBOR });
        const _state = new GomokuState({
            context: _context,
            scoreMap: props.scoreMap ?? createScoreMap(_context.MAX_ADJACENT),
        });
        const _MIN_DEPTH = Math.max(1, Math.round(MIN_DEPTH));
        const _MAX_DEPTH = Math.max(_MIN_DEPTH, Math.round(MAX_DEPTH));
        const _POSSIBILITY_SEARCH_EQUIV_CANDIDATE = Math.min(1, Math.max(0, POSSIBILITY_SEARCH_EQUIV_CANDIDATE));
        const _queues = new Array(_MAX_DEPTH)
            .fill([])
            .map(() => priorityQueue.createPriorityQueue((x, y) => x.score - y.score));
        this.MIN_DEPTH = _MIN_DEPTH;
        this.MAX_DEPTH = _MAX_DEPTH;
        this.POSSIBILITY_SEARCH_EQUIV_CANDIDATE = _POSSIBILITY_SEARCH_EQUIV_CANDIDATE;
        this.context = _context;
        this.state = _state;
        this._cache = new GomokuStateCache(BigInt(_context.TOTAL_POS));
        this._queues = _queues;
        this._mainPlayerId = -1;
        this._bestMoveId = -1;
    }
    init(pieces) {
        this.context.init(pieces);
        this.state.init(pieces);
        this._cache.clear();
    }
    forward(r, c, playerId) {
        if (this.context.isValidPos(r, c)) {
            const posId = this.context.idx(r, c);
            this._forward(posId, playerId);
        }
    }
    revert(r, c) {
        if (this.context.isValidPos(r, c)) {
            const posId = this.context.idx(r, c);
            this._revert(posId);
        }
    }
    minimaxSearch(nextPlayer) {
        if (this.state.isFinal())
            return [-1, -1];
        this._cache.clear();
        this._bestMoveId = -1;
        this._mainPlayerId = nextPlayer;
        this._alphaBeta(nextPlayer, Number.NEGATIVE_INFINITY, Number.POSITIVE_INFINITY, 0, this._cache.INITIAL_STATE);
        if (this._bestMoveId < 0) {
            throw new Error('Oops! Something must be wrong, cannot find a valid moving strategy');
        }
        const [r, c] = this.context.revIdx(this._bestMoveId);
        return [r, c];
    }
    _alphaBeta(player, alpha, beta, cur, prevState) {
        const { MIN_DEPTH, MAX_DEPTH, state, _cache, _mainPlayerId } = this;
        if (state.isWin(_mainPlayerId))
            return Number.POSITIVE_INFINITY;
        if (state.isWin(_mainPlayerId ^ 1))
            return Number.NEGATIVE_INFINITY;
        if (cur === MAX_DEPTH || state.isDraw())
            return state.score(player ^ 1, _mainPlayerId);
        const Q = this._queues[cur];
        let _size = state.expand(player, tmpCandidates);
        Q.init(tmpCandidates, 0, _size);
        const firstCandidate = Q.top();
        const MAX_CANDIDATE_SCORE = firstCandidate.score;
        const POSS_SEARCH_EQUIV = this.POSSIBILITY_SEARCH_EQUIV_CANDIDATE;
        const shouldCache = cur > 1 && cur + 1 < MAX_DEPTH;
        if (cur === 0)
            this._bestMoveId = firstCandidate.id;
        if (cur >= MIN_DEPTH && _size > 8)
            _size = 8;
        for (let i = 0, prevCandidateScore = -1, possibility = POSS_SEARCH_EQUIV; i < _size; ++i) {
            let candidate = Q.dequeue();
            if (candidate.score === prevCandidateScore) {
                for (; Math.random() >= possibility; possibility *= POSS_SEARCH_EQUIV) {
                    i += 1;
                    if (i === _size)
                        break;
                    candidate = Q.dequeue();
                }
                if (i === _size)
                    break;
            }
            if (prevCandidateScore !== candidate.score) {
                prevCandidateScore = candidate.score;
                possibility = POSS_SEARCH_EQUIV;
            }
            const posId = candidate.id;
            const nextState = _cache.calcNextState(cur, prevState, posId);
            let gamma = _cache.get(nextState);
            if (gamma === undefined) {
                if (candidate.score * 16 < MAX_CANDIDATE_SCORE)
                    continue;
                this._forward(posId, player);
                gamma = this._alphaBeta(player ^ 1, alpha, beta, cur + 1, nextState);
                this._revert(posId);
                if (shouldCache)
                    _cache.set(nextState, gamma);
            }
            if (player === _mainPlayerId) {
                if (alpha < gamma) {
                    alpha = gamma;
                    if (cur === 0)
                        this._bestMoveId = posId;
                }
            }
            else {
                if (beta > gamma) {
                    beta = gamma;
                }
            }
            if (beta <= alpha)
                break;
        }
        return player === _mainPlayerId ? alpha : beta;
    }
    _forward(posId, playerId) {
        this.context.forward(posId, playerId);
        this.state.forward(posId);
    }
    _revert(posId) {
        this.context.revert(posId);
        this.state.revert(posId);
    }
}

exports.GomokuContext = GomokuContext;
exports.GomokuCountMap = GomokuCountMap;
exports.GomokuDirectionTypeBitset = GomokuDirectionTypeBitset;
exports.GomokuDirectionTypes = GomokuDirectionTypes;
exports.GomokuDirections = GomokuDirections;
exports.GomokuSolution = GomokuSolution;
exports.GomokuState = GomokuState;
exports.GomokuStateCache = GomokuStateCache;
exports.GomokuStateCompressor = GomokuStateCompressor;
exports.createScoreMap = createScoreMap;
